# Copyright 2016 Jochen Fabricius
# Distribution and modification is encouraged

namespace eval filter {
  # filter coefficients
  # format:
  # Type {Order {frequency_factor Q} Order {....}} Type {...}
  # Q = {-} for single order filters (important!)
  variable coeffs
  array set coeffs {
    Linkwitz {
      1 {1.000 -}
      2 {1.000 0.50}
      3 {1.000 -    1.000 1.00}
      4 {1.000 0.71 1.000 0.71}
      5 {1.000 -    1.000 0.62 1.000 1.62}
      6 {1.000 0.52 1.000 0.71 1.000 1.93}
    }
    Butterworth {
      1 {1.000 -}
      2 {1.000 0.71}
      3 {1.000 -    1.000 1.0}
      4 {1.000 0.54 1.000 1.31}
      5 {1.000 -    1.000 0.62 1.000 1.62}
      6 {1.000 0.52 1.000 0.71 1.000 1.93}
    }
    Bessel {
      1 {1.000 -}
      2 {1.274 0.58}
      3 {1.323 -    1.448 0.69}
      4 {1.430 0.52 1.603 0.81}
      5 {1.502 -    1.556 0.56 1.755 0.92}
      6 {1.604 0.51 1.689 0.61 1.905 1.02}
    }
    Chebychev {
      1 {1.000 -}
      2 {0.863 0.96}
      3 {0.452 -    0.911 2.02}
      4 {0.502 0.80 0.943 3.56}
      5 {0.280 -    0.634 1.41 0.961 5.56}
      6 {0.342 0.76 0.723 2.20 0.963 8.00}
    }
  }

  proc coefficients {Shape Order} {
    variable coeffs
    if {[info exists coeffs($Shape)]} {
      array set shape $coeffs($Shape)
      if {[info exists shape($Order)]} {
              return $shape($Order)
      }
    }
    return {}
  }
}

# if $coeffs starts with a bare word like "Butterworth" then get the coefficients from the table
# otherwise, it must be a coefficient list comprising elements of the form.
# a) a 2 elements sublist {f q} like in the table above, where $q may be a "-", 
# and f is a relative frequency factor; the position of the zeros depends on the type of filter 
# (e. g. for a lowpass they are at infinity). A notch can't be formed with this format
# b) a 2 elements sublist, comprising again of a 2 elements sublist, {{fp qp} {fz qz}}, 
# where fp and qp denote the poles and fz qz denotes the zeros. Both fp and fz are relative frequency factors, 
# while qp and qz may be a single "-" meaning a real valued pole or zero, resp. 
# The latter form is needed for notches and shelves
# Examples: 
# 1) a 2nd order Linkwitz lowpass at 1000 Hz
# :filter #1 lowpass 1000 {Linkwitz 2}
# 2) a 3rd order highpass with custom coefficients
# :filter #1 highpass 1000 {{1.0 0.8} {0.8 -}}
# 3) a 2nd order notch
# :filter #1 notch 1000 {{{1.0 0.8} {1.0 1.2}}}
proc :filter {ID type freq {coeffs {Butterworth 2}}} {
  if {[catch {isID $ID} id]} {
    return -code error $id
  }
  set ir [getProp $id IR]
  if {[catch {esweep::size -obj $ir}]} {
    return -code error "No impulse response for ID $ID available"
  }

  set Fs [esweep::samplerate -obj $ir]
  # get coefficient list
  if {[llength [lindex $coeffs 0 0]] == 1} {
    if {[catch {expr {int([lindex $coeffs 0 0])}}]} {
      # first element is not a number, try to get the coefficients from the table
      lassign $coeffs shape order
      if {[set poles [filter::coefficients $shape $order]] eq {}} {
        return -code error "Invalid shape $shape or order $order"
      }
      DEBUGPRINT "Poles by shape $shape: $poles"
      DEBUGPRINT {No zeros given}
    } else {
      # case a), zeros in the origin
      set poles [join $coeffs { }]
      if {[llength $poles] % 2 != 0} {
        return -code error "Coefficient table must have an even number of elements"
      }
      DEBUGPRINT "Poles: $poles"
      DEBUGPRINT {No zeros given}
    }
  } elseif {[llength [lindex $coeffs 0 0]] == 2} {
    # case b), zeros somewhere in the z-plane
    foreach {p z} {*}$coeffs {
      lappend poles {*}$p
      lappend zeros {*}$z
    }
    DEBUGPRINT "Poles: $poles"
    DEBUGPRINT "Zeros: $zeros"
  } else {
    return -code error "Unknown coefficient format"
  }

  switch -nocase -- $type {
    lowpass -
    LP {
      foreach {f q} $poles {
        if {$q eq {-}} {
          # filter of odd order
          # we must use an integrator
          lappend filter \
            [esweep::createFilterFromCoeff -samplerate $Fs \
              -type integrator \
              -gain 1.0 \
              -Fp [expr {$f*$freq}]]
        } else {
          # even order
          lappend filter \
            [esweep::createFilterFromCoeff -samplerate $Fs \
              -type lowpass \
              -gain 1.0 \
              -Fp [expr {$f*$freq}] -Qp $q]
        }
      }
    }
    highpass -
    HP {
      foreach {f q} $poles {
        if {$q eq {-}} {
          # filter of odd order
          # we must use an integrator
          lappend filter \
            [esweep::createFilterFromCoeff -samplerate $Fs \
              -type differentiator \
              -gain 1.0 \
              -Fp [expr {$f*$freq}]]
        } else {
          # even order
          lappend filter \
            [esweep::createFilterFromCoeff -samplerate $Fs \
              -type highpass \
              -gain 1.0 \
              -Fp [expr {$f*$freq}] -Qp $q]
        }
      }
    }
    bandpass -
    BP {
      foreach {f q} $poles {
        if {$q eq {-}} {
          # filter of odd order
          # not possible, ignore
        } else {
          # even order
          lappend filter \
            [esweep::createFilterFromCoeff -samplerate $Fs \
              -type bandpass \
              -gain 1.0 \
              -Fp [expr {$f*$freq}] -Qp $q]
        }
      }
    }
    bandstop -
    BS {
      foreach {f q} $poles {
        if {$q eq {-}} {
          # filter of odd order
          # not possible, ignore
        } else {
          # even order
          lappend filter \
            [esweep::createFilterFromCoeff -samplerate $Fs \
              -type bandstop \
              -gain 1.0 \
              -Fp [expr {$f*$freq}] -Qp $q]
        }
      }
    }
    notch {
      foreach {fp qp} $poles {fz qz} $zeros {
        if {($qp eq {-}) || ($qz eq {-})} {
          # filter of odd order
          # not possible, ignore
        } else {
          # even order
          lappend filter \
            [esweep::createFilterFromCoeff -samplerate $Fs \
              -type notch \
              -gain 1.0 \
              -Fp [expr {$fp*$freq}] -Qp $qp -Fz [expr {$fz*$freq}] -Qz $qz]
        }
      }
    }
    default {
      return -code error "Unknown filter type $type"
    }
  }
  # apply filter
  foreach biquad $filter {
    set chain $biquad
    break
  }
  foreach biquad [lrange $filter 1 end] {
    set chain [esweep::appendFilter -filter $chain -append $biquad]
  }
  esweep::filter -signal ir -filter $chain 

  ::overlay::modify $id [list FR {}]
  ::overlay::modify $id [list IR $ir]
  return $ID
}

proc weighting {FS type} {
  switch -- $type {
    A {
      set filter [esweep::createFilterFromCoeff -type highpass -samplerate $FS \
        -Fp 20.6 -Qp 0.5]
      set filter [esweep::appendFilter -filter $filter -append \
        [esweep::createFilterFromCoeff -type highpass -samplerate $FS -Fp 12200 -Qp 0.5]]
      set filter [esweep::appendFilter -filter $filter -append \
        [esweep::createFilterFromCoeff -type integrator -samplerate $FS -Fp 107.7 -Qp 0.5]]
      set filter [esweep::appendFilter -filter $filter -append \
        [esweep::createFilterFromCoeff -type integrator -samplerate $FS -Fp 737.9 -Qp 0.5]]
      set sc 3.890e3
    }
    B {
      set filter [esweep::createFilterFromCoeff -type bandpass -samplerate $FS \
        -Fp 20.6 -Qp 0.5]
      set filter [esweep::appendFilter -filter $filter -append \
        [esweep::createFilterFromCoeff -type highpass -samplerate $FS -Fp 12200 -Qp 0.5]]
      set filter [esweep::appendFilter -filter $filter -append \
        [esweep::createFilterFromCoeff -type integrator -samplerate $FS -Fp 158.5 -Qp 0.5]]
      set sc 7.674e4
    }
    C {
      set filter [esweep::createFilterFromCoeff -type highpass -samplerate $FS \
        -Fp 20.6 -Qp 0.5]
      set filter [esweep::appendFilter -filter $filter -append \
        [esweep::createFilterFromCoeff -type lowpass -samplerate $FS -Fp 12200 -Qp 0.5]]
      set sc 1
    }
    D {
      set filter [esweep::createFilterFromCoeff -type linkwitz -samplerate $FS \
        -Fp 3136.5 -Qp 0.92 -Fz 1018.8 -Qz 0.9]
      set filter [esweep::appendFilter -filter $filter -append \
        [esweep::createFilterFromCoeff -type integrator -samplerate $FS -Fp 282.7 -Qp 0.5]]
      set filter [esweep::appendFilter -filter $filter -append \
        [esweep::createFilterFromCoeff -type integrator -samplerate $FS -Fp 1160 -Qp 0.5]]
      set filter [esweep::appendFilter -filter $filter -append \
        [esweep::createFilterFromCoeff -type differentiator -samplerate $FS -Fp 20e3 -Qp 0.5]]
      set sc 2.6e3
    }
    ITU468 {
      set f 8000
      set q 1
      set filter [esweep::createFilterFromCoeff -type lowpass -samplerate $FS \
        -Fp $f -Qp $q]
      set filter [esweep::appendFilter -filter $filter -append \
        [esweep::createFilterFromCoeff -type lowpass -samplerate $FS -Fp $f -Qp $q]]
      set filter [esweep::appendFilter -filter $filter -append \
        [esweep::createFilterFromCoeff -type integrator -samplerate $FS -Fp $f]]
      set filter [esweep::appendFilter -filter $filter -append \
        [esweep::createFilterFromCoeff -type differentiator -samplerate $FS -Fp 20e3]]
      set sc 56.2
    }
    default {
      return
    }
  }
  return [list $sc $filter]
}

proc :resample {ID FS_out {cutoff_factor 0.9} {filter_length 128}} {
  if {[catch {isID $ID} id]} {
    return -code error $id
  }
  set ir [getProp $id IR]
  if {[catch {esweep::size -obj $ir}]} {
    return -code error "No impulse response for ID $ID available"
  }
  set ir [::audio::resample $ir $FS_out $cutoff_factor $filter_length]
  ::overlay::modify $id [list IR $ir]
  return $ID
}

# simply create a dirac function and add it to the overlays
# length is that of the current gate
proc :dirac {{delay 0}} {
  set FS [setup cget Audio,Samplerate]
  set gate [setup cget Processing,Gate]
  lassign [lmap t $gate {expr {$t > 0.0 ? int(0.5+$t*$FS/1000.0) : 0}}] gate_start gate_stop
  set gate_length [expr {$gate_stop-$gate_start}]

  set dirac [esweep::create -type {wave} -size $gate_length -samplerate $FS]
  esweep::generate dirac -obj dirac -delay $delay

  # add to overlays
  set props [list Name Dirac Color red \
                SigType {Generated} \
                SweepRate -1 \
                Gate [list $gate_start $gate_stop] \
                Window [list {rect} 100] \
                Mode {Generated} \
                OutLvl 1 \
                IR $dirac \
                Mic {} \
                MicExp {} \
                Show {yes}]

  set id [::overlay::add $props]
  # force display update
  ::overlay::modify $id [list IR $dirac]

  return #$id
}

